@@ -1,8 +1,5 @@
 #!/usr/bin/env python
 
-import sys
-import os
-import glob
 import optparse
 import gtk
 import tl.eggdeps.graph
@@ -23,13 +20,11 @@ class ExtraDiscerningGraph(tl.eggdeps.graph.Graph):
     form 'project_name[extra]'.  As a result, if a package B requres package X
     with [extra1, extra2], you'll see three names in the specs list::
 
-        graph['B'].deps == {'X': set(),
-                            'X[extra1]': set(),
-                            'X[extra2]': set()}
+        graph['B'] == {'X': set(), 'X[extra1]': set(), 'X[extra2]': set()}
 
-    Since deps is also used for graph traversal, I override __getitem__()
-    so that graph['X[extra1'] is the same thing as graph['X'].  It is perhaps
-    not very nice, but it seems to work.
+    Since the node dict keys are also used for graph traversal, I override
+    __getitem__() so that graph['X[extra1'] is the same thing as graph['X'].
+    It is perhaps not very nice, but it seems to work.
     """
 
     def __getitem__(self, name):
@@ -47,19 +42,124 @@ class ExtraDiscerningGraph(tl.eggdeps.graph.Graph):
         return names
 
 
+class Node(set):
+    """A generic graph node."""
+
+    def __init__(self, name, **kw):
+        self.name = name
+        self._kwargs = kw
+        self.__dict__.update(kw)
+
+    def get_kwargs(self):
+        return self._kwargs
+
+
+class Graph(dict):
+    """A generic directed graph.
+
+    Construct the graph by calling ``add_node`` and ``add_edge``.
+
+    Access graph nodes by name via ``graph[node_name]``.  Get an adjacency
+    list by ``list(graph[node_name])``.
+    """
+
+    def __init__(self): # disallow constructor arguments
+        pass
+
+    def add_node(self, node_id, **kw):
+        if node_id in self:
+            raise ValueError('node already exists: %s' % node_id)
+        self[node_id] = Node(node_id, **kw)
+
+    def add_edge(self, node_id, child_id):
+        if node_id not in self:
+            self.add_node(node_id, False)
+        self[node_id].add(child_id)
+
+    def reversed(self):
+        clone = self.__class__()
+        for node_id, node in self.iteritems():
+            clone.add_node(node_id, **node.get_kwargs())
+        for node_id, node in self.iteritems():
+            for child_id in node:
+                clone.add_edge(child_id, node_id)
+        return clone
+
+
+def DependencyGraph(specifications=None, **kw):
+    """Construct a package dependency graph.
+
+    The nodes of this graph are packages ("distributions" in
+    setuptools-speak) and extras (unlike tl.eggdeps, which uses only
+    distributions as nodes).
+    """
+    # we could use tl.eggdeps.graph.Graph() directly if we can be sure we've
+    # got a recent-enough version that supports iter_deps_with_extras, i.e.
+    # r1532 or later.
+    egggraph = ExtraDiscerningGraph(**kw)
+    if specifications is None:
+        egggraph.from_working_set()
+    else:
+        egggraph.from_specifications(*specifications)
+    graph = Graph()
+    for node_id, node in egggraph.iteritems():
+        graph.add_node(node_id, compatible=node.compatible)
+        if node.dist:
+            # tl.eggdeps did not have the information we need before
+            # r1527, so we have to fall back to node.dist.extras.  It's
+            # not very accurate -- we should ignore extras that weren't
+            # required by any package, since their dependency information is
+            # not loaded anyway.
+            extras = getattr(node, 'extras_used', node.dist.extras)
+            for extra in sorted(extras):
+                graph.add_node(node_id + '[' + extra +']',
+                              compatible=node.compatible)
+    for node_id, node in egggraph.iteritems():
+        if hasattr(node, 'iter_deps_with_extras'):
+            # available since tl.eggdeps r1532.
+            for child_id, dep_extra, extras in sorted(
+                                                node.iter_deps_with_extras()):
+                if extras:
+                    for extra in sorted(extras):
+                        graph.add_edge(node_id + '[' + extra + ']', child_id)
+                else:
+                    graph.add_edge(node_id, child_id)
+                if dep_extra:
+                    child_id += '[' + dep_extra + ']'
+                    if extras:
+                        for extra in sorted(extras):
+                            graph.add_edge(node_id + '[' + extra + ']', child_id)
+                    else:
+                        graph.add_edge(node_id, child_id)
+        else:
+            # the API of tl.eggdeps changed in r1524; before it we used
+            # iteritems() to get sets of extras, now we'd be getting dicts
+            # of some kind; to get the old data structure we should be
+            # using iter_deps().
+            iter_deps = getattr(node, 'iter_deps', node.iteritems)
+            for child_id, extras in sorted(iter_deps()):
+                if extras:
+                    for extra in sorted(extras):
+                        graph.add_edge(node_id + '[' + extra + ']', child_id)
+                else:
+                    graph.add_edge(node_id, child_id)
+    return graph
+
+
 class MyTreeModel(gtk.GenericTreeModel):
     """Represent a graph as a tree with duplicated nodes (and subtrees).
 
-    Handles cycles correctly.
+    Handles cycles correctly (mostly).
     """
 
-    def __init__(self, graph, reverse=False):
+    def __init__(self, graph, reverse=False,
+                 show_all_packages_at_top_level=False):
         gtk.GenericTreeModel.__init__(self)
-        self.build_graph(graph, reverse)
+        self.build_graph(graph, reverse, show_all_packages_at_top_level)
         self.calc_depths()
         self.calc_dep_sizes()
 
-    def build_graph(self, graph, reverse):
+    def build_graph(self, graph, reverse, show_all_packages_at_top_level):
         self.reverse = reverse
         self.children = {None: []}
         self.node_name = {}
@@ -69,25 +169,17 @@ class MyTreeModel(gtk.GenericTreeModel):
             if not node.compatible:
                 node_name = '(%s)' % node_name
             self.add_node(node_id, node_name, node.compatible)
-            if node.dist:
-                for extra in sorted(node.dist.extras):
-                    self.add_node(node_id + '[' + extra +']',
-                                  node_name + ' [' + extra + ']',
-                                  node.compatible)
         for node_id, node in graph.iteritems():
-            for child_id, extras in sorted(node.items()):
-                if extras:
-                    for extra in sorted(extras):
-                        self.add_edge(node_id + '[' + extra + ']', child_id)
-                else:
-                    self.add_edge(node_id, child_id)
+            for child_id in sorted(node):
+                self.add_edge(node_id, child_id)
         roots = set(self.children)
         roots.remove(None)
-        # XXX: if there are loops, they won't be reachable from roots!
-        for children in self.children.values():
-            for child_id in children:
-                if child_id in roots:
-                    roots.remove(child_id)
+        if not show_all_packages_at_top_level:
+            # XXX: if there are loops, they won't be reachable from roots!
+            for children in self.children.values():
+                for child_id in children:
+                    if child_id in roots:
+                        roots.remove(child_id)
         self.children[None] = sorted(roots)
 
     def add_node(self, node_id, node_name, available):
@@ -249,14 +341,17 @@ class MainWindow(object):
     def load_tree(self, opts, specs):
         def show(name):
             return name not in opts.ignore
-        graph = ExtraDiscerningGraph(show=show)
+        if not specs:
+            specs = None # convert [] to None
+        graph = DependencyGraph(specs, show=show)
         if specs:
             self.window.set_title("Egg dependencies for %s" % " ".join(specs))
-            graph.from_specifications(*specs)
+            show_all_packages_at_top_level = False
         else:
             self.window.set_title("Egg dependencies of the working set")
-            graph.from_working_set()
-        self.model = MyTreeModel(graph, opts.reverse)
+            show_all_packages_at_top_level = True
+        self.model = MyTreeModel(graph, opts.reverse,
+                                 show_all_packages_at_top_level)
         self.tree.set_model(self.model)
 
 
